@@ -23,8 +23,14 @@
    Bastien Nocera <hadess@hadess.net>
 */
 
+#ifdef HAVE_CONFIG_H
 #include "config.h"
+#endif
 
+#include <glib.h>
+#include <glib/gi18n.h>
+
+#include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/types.h>
@@ -35,6 +41,10 @@
 #include <fcntl.h>
 #include <math.h>
 
+#ifdef USE_HAL
+#include <libhal.h>
+#endif /* USE_HAL */
+
 #ifdef __linux__
 #include <scsi/scsi.h>
 #include <scsi/sg.h>
@@ -46,9 +56,6 @@
 #include <camlib.h>
 #endif /* __FreeBSD__ */
 
-#include <glib.h>
-#include <libgnome/gnome-i18n.h>
-
 #include "cd-drive.h"
 
 static struct {
@@ -59,13 +66,159 @@ static struct {
 	gboolean can_write_dvdram;
 } recorder_whitelist[] = {
 	{ "IOMEGA - CDRW9602EXT-B", TRUE, TRUE, FALSE, FALSE },
+	{ "SONY - CD-R   CDU948S",  TRUE, FALSE, FALSE, FALSE },
 };
 
-#if defined(__linux__) || defined(__FreeBSD__)
+/* Utility functions, be careful to have a match with what's use in the
+ * different bits of code */
+static char *
+cdrecord_get_stdout_for_id (char *id)
+{
+	int max_speed, i;
+	const char *argv[20]; /* Shouldn't need more than 20 arguments */
+	char *dev_str, *stdout_data;
+
+	max_speed = -1;
+
+	i = 0;
+	argv[i++] = "cdrecord";
+	argv[i++] = "-prcap";
+	dev_str = g_strdup_printf ("dev=%s", id);
+	argv[i++] = dev_str;
+	argv[i++] = NULL;
+
+	if (g_spawn_sync (NULL,
+				(char **)argv,
+				NULL,
+				G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL,
+				NULL, NULL,
+				&stdout_data,
+				NULL,
+				NULL,
+				NULL)) {
+		g_free (dev_str);
+		return stdout_data;
+	}
+
+	g_free (dev_str);
+	return NULL;
+}
+
+#if !defined (__linux__)
+static void
+get_cd_properties (char *id, int *max_rd_speed, int *max_wr_speed,
+	CDDriveType *type)
+{
+	char *stdout_data, *rd_speed, *wr_speed, *drive_cap;
+
+	*max_rd_speed = -1;
+	*max_wr_speed = -1;
+	*type = 0;
+
+	stdout_data = cdrecord_get_stdout_for_id (id);
+	if (stdout_data == NULL) {
+		return;
+	}
+
+	wr_speed = strstr (stdout_data, "Maximum write speed:");
+	if (wr_speed != NULL) {
+		char *tok;
+		wr_speed += strlen ("Maximum write speed:");
+		for (tok = strtok (wr_speed, " (),\t\n");
+				tok && strcmp (tok, "CD");
+				tok = strtok (NULL, " (),\t\n")) {}
+		tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
+		*max_wr_speed = atol (tok);
+	} else {
+		wr_speed = strstr (stdout_data, "Maximum write speed in kB/s:");
+		if (wr_speed != NULL) {
+			char *tok;
+			wr_speed += strlen ("Maximum write speed in kB/s:");
+			for (tok = strtok (wr_speed, " (),\t\n");
+					tok && strcmp (tok, "CD");
+					tok = strtok (NULL, " (),\t\n")) {}
+			tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
+			*max_wr_speed = atol (tok);
+		}
+	}
+
+	rd_speed = strstr (stdout_data, "Maximum read  speed:");
+	if (rd_speed != NULL) {
+		char *tok;
+		rd_speed += strlen ("Maximum read  speed:");
+		for (tok = strtok (rd_speed, " (),\t\n"); 
+				tok && strcmp (tok, "CD"); 
+				tok = strtok (NULL, " (),\t\n")) {}
+		tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
+		*max_rd_speed = atol (tok);
+	} else {
+		rd_speed = strstr (stdout_data, "Maximum read  speed in kB/s:");
+		if (rd_speed != NULL) {
+			char *tok;
+			rd_speed += strlen ("Maximum read  speed in kB/s:");
+			for (tok = strtok (rd_speed, " (),\t\n");
+					tok && strcmp (tok, "CD");
+					tok = strtok (NULL, " (),\t\n")) {}
+			tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
+			*max_rd_speed = atol (tok);
+		}
+	}
+	drive_cap = strstr (stdout_data, "Does write DVD-RAM media");
+	if (drive_cap != NULL) {
+		*type |= CDDRIVE_TYPE_DVD_RAM_RECORDER;
+	}
+	drive_cap = strstr (stdout_data, "Does read DVD-R media");
+	if (drive_cap != NULL) {
+		*type |= CDDRIVE_TYPE_DVD_RW_RECORDER;
+	}
+	drive_cap = strstr (stdout_data, "Does read DVD-ROM media");
+	if (drive_cap != NULL) {
+		*type |= CDDRIVE_TYPE_DVD_DRIVE;
+	}
+	drive_cap = strstr (stdout_data, "Does write CD-RW media");
+	if (drive_cap != NULL) {
+		*type |= CDDRIVE_TYPE_CDRW_RECORDER;
+	}
+	drive_cap = strstr (stdout_data, "Does write CD-R media");
+	if (drive_cap != NULL) {
+		*type |= CDDRIVE_TYPE_CD_RECORDER;
+	}
+	drive_cap = strstr (stdout_data, "Does read CD-R media");
+	if (drive_cap != NULL) {
+		*type |= CDDRIVE_TYPE_CD_DRIVE;
+	}
+	g_free (stdout_data);
+}
+#endif /* !__linux__ */
+
+#if !(defined (__linux__)) || (defined (__linux__) && (defined (USE_HAL)))
+static void
+add_whitelist (CDDrive *cdrom)
+{
+	guint i;
+
+	for (i = 0; i < G_N_ELEMENTS (recorder_whitelist); i++) {
+		if (!strcmp (cdrom->display_name, recorder_whitelist[i].name)) {
+			if (recorder_whitelist[i].can_write_cdr) {
+				cdrom->type |= CDDRIVE_TYPE_CD_RECORDER;
+			}
+			if (recorder_whitelist[i].can_write_cdrw) {
+				cdrom->type |= CDDRIVE_TYPE_CDRW_RECORDER;
+			}
+			if (recorder_whitelist[i].can_write_dvdr) {
+				cdrom->type |= CDDRIVE_TYPE_DVD_RW_RECORDER;
+			}
+			if (recorder_whitelist[i].can_write_dvdram) {
+				cdrom->type |= CDDRIVE_TYPE_DVD_RAM_RECORDER;
+			}
+		}
+	}
+}
+#endif /* !__linux__ */
 
 /* For dvd_plus_rw_utils.cpp */
 int get_dvd_r_rw_profile (const char *name);
-int get_mmc_profile (void *fd);
+int get_mmc_profile (int fd);
 
 static void
 add_dvd_plus (CDDrive *cdrom)
@@ -88,18 +241,23 @@ add_dvd_plus (CDDrive *cdrom)
 	}
 }
 
-static CDMediaType
-linux_bsd_media_type (const char *device)
+CDMediaType
+cd_drive_get_media_type_from_path (const char *device)
 {
 	int fd;
 	int mmc_profile;
 
-	fd = open (device, O_RDONLY);
+	g_return_val_if_fail (device != NULL, CD_MEDIA_TYPE_ERROR);
+
+	fd = open (device, O_RDONLY|O_EXCL|O_NONBLOCK);
 	if (fd < 0) {
+		if (errno == EBUSY) {
+			return CD_MEDIA_TYPE_BUSY;
+		}
 		return CD_MEDIA_TYPE_ERROR;
 	}
 
-	mmc_profile = get_mmc_profile ((void *)&fd);
+	mmc_profile = get_mmc_profile (fd);
 	close (fd);
 
 	switch (mmc_profile) {
@@ -129,29 +287,149 @@ linux_bsd_media_type (const char *device)
 	}
 }
 
+CDMediaType
+cd_drive_get_media_type (CDDrive *cdrom)
+{
+	g_return_val_if_fail (cdrom != NULL, CD_MEDIA_TYPE_ERROR);
+
+	return cd_drive_get_media_type_from_path (cdrom->device);
+}
+
+#ifdef USE_HAL
+
+#define GET_BOOL_PROP(x) (hal_device_property_exists (ctx, device_names[i], x) && hal_device_get_property_bool (ctx, device_names[i], x))
+#define CD_ROM_SPEED 176
+
+static GList *
+hal_scan (gboolean recorder_only)
+{
+	GList *cdroms = NULL;
+	LibHalFunctions hal_functions = {
+		NULL, /* mainloop integration */
+		NULL, /* device_added */
+		NULL, /* device_removed */
+		NULL, /* device_new_capability */
+		NULL, /* property_modified */
+		NULL, /* device_condition */
+	};
+	int i;
+	int num_devices;
+	char** device_names;
+	LibHalContext *ctx;
+
+	ctx = hal_initialize (&hal_functions, FALSE);
+	if (ctx == NULL) {
+		return NULL;
+	}
+
+	device_names = hal_get_all_devices (ctx, &num_devices);
+	if (device_names == NULL)
+	{
+		hal_shutdown (ctx);
+		return NULL;
+	}
+
+	for (i = 0; i < num_devices; i++)
+	{
+		CDDrive *cdrom;
+		char *string;
+		gboolean is_cdr;
+
+		/* Is it a removable drive? */
+		if (!GET_BOOL_PROP ("storage.removable")) {
+			continue;
+		}
+
+		/* Is it a CD drive? */
+		string = hal_device_get_property_string (ctx, device_names[i],
+				"storage.media");
+		if (string == NULL || strcmp (string, "cdrom") != 0) {
+			g_free (string);
+			continue;
+		}
+		g_free (string);
+
+		/* Is it the CD-drive, or a volume on the media? */
+		if (GET_BOOL_PROP ("block.is_volume") != FALSE) {
+			continue;
+		}
+
+		/* Is it a CD burner? */
+		is_cdr = GET_BOOL_PROP ("storage.cdr");
+
+		cdrom = g_new0 (CDDrive, 1);
+		cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
+		if (is_cdr != FALSE) {
+			cdrom->type |= CDDRIVE_TYPE_CD_RECORDER;
+		}
+		if (GET_BOOL_PROP ("storage.cdrw")) {
+			cdrom->type |= CDDRIVE_TYPE_CDRW_RECORDER;
+		}
+		if (GET_BOOL_PROP ("storage.dvd")) {
+			cdrom->type |= CDDRIVE_TYPE_DVD_DRIVE;
+
+			if (GET_BOOL_PROP ("storage.dvdram")) {
+				cdrom->type |= CDDRIVE_TYPE_DVD_RAM_RECORDER;
+			}
+			if (GET_BOOL_PROP ("storage.dvdr")) {
+				cdrom->type |= CDDRIVE_TYPE_DVD_RW_RECORDER;
+			}
+			if (GET_BOOL_PROP ("storage.dvd")) {
+				cdrom->type |= CDDRIVE_TYPE_DVD_DRIVE;
+			}
+			if (GET_BOOL_PROP ("storage.dvdplusr")) {
+				cdrom->type |= CDDRIVE_TYPE_DVD_PLUS_R_RECORDER;
+			}
+			if (GET_BOOL_PROP ("storage.dvdplusrw")) {
+				cdrom->type |= CDDRIVE_TYPE_DVD_PLUS_RW_RECORDER;
+			}
+		}
+
+		cdrom->device = hal_device_get_property_string (ctx,
+				device_names[i], "block.device");
+		cdrom->cdrecord_id = g_strdup (cdrom->device);
+
+		string = hal_device_get_property_string (ctx,
+				device_names[i], "storage.model");
+		if (string != NULL) {
+			cdrom->display_name = string;
+		} else {
+			cdrom->display_name = g_strdup_printf ("Unnamed CD-ROM (%s)", cdrom->device);
+		}
+
+		cdrom->max_speed_read = hal_device_get_property_int
+			(ctx, device_names[i], "storage.cdrom.read_speed")
+			/ CD_ROM_SPEED;
+
+		if (hal_device_property_exists (ctx, device_names[i], "storage.cdrom.write_speed")) {
+			cdrom->max_speed_write = hal_device_get_property_int
+				(ctx, device_names[i],
+				 "storage.cdrom.write_speed")
+				/ CD_ROM_SPEED;
+		}
+
+		add_whitelist (cdrom);
+
+		if (recorder_only && !(cdrom->type & CDDRIVE_TYPE_CD_RECORDER)) {
+			cd_drive_free (cdrom);
+		} else {
+			cdroms = g_list_prepend (cdroms, cdrom);
+		}
+	}
+
+	cdroms = g_list_reverse (cdroms);
+
+	return cdroms;
+}
+#endif /* USE_HAL */
+
+#if defined(__linux__) || defined(__FreeBSD__)
+
+
+
 #endif /* __linux__ || __FreeBSD__ */
 
-#ifdef __linux__
-
-#ifdef USE_STABLE_LIBGLIB
-static gboolean
-g_str_has_prefix (gchar *haystack, gchar *needle)
-{
-	if (haystack == NULL && needle == NULL) {
-		return TRUE;
-	}
-
-	if (haystack == NULL || needle == NULL) {
-		return FALSE;
-	}
-
-	if (strncmp (haystack, needle, strlen (needle)) == 0) {
-		return TRUE;
-	}
-
-	return FALSE;
-}
-#endif /* USE_STABLE_LIBGLIB */
+#if defined (__linux__)
 
 static char **
 read_lines (char *filename)
@@ -210,7 +488,7 @@ struct cdrom_unit {
 static char *cdrom_get_name (struct cdrom_unit *cdrom, struct scsi_unit *scsi_units, int n_scsi_units);
 
 static void
-add_whitelist (struct cdrom_unit *cdrom_s,
+linux_add_whitelist (struct cdrom_unit *cdrom_s,
 	       struct scsi_unit *scsi_units, int n_scsi_units)
 {
 	guint i;
@@ -241,7 +519,9 @@ get_scsi_units (char **device_str, char **devices, struct scsi_unit *scsi_units)
 	int scsi_id, scsi_lun, scsi_type;
 	int i, j;
 
-	for (i = 0, j = 0; device_str[i] != NULL && devices[i] != NULL; i++) {
+	j = 0;
+
+	for (i = 0; device_str[i] != NULL && devices[i] != NULL; i++) {
 		if (strcmp (device_str[i], "<no active device>") == 0) {
 			continue;
 		}
@@ -272,7 +552,6 @@ get_scsi_units (char **device_str, char **devices, struct scsi_unit *scsi_units)
 			j++;
 		}
 	}
-	
 }
 
 static int
@@ -306,17 +585,22 @@ get_cd_scsi_id (const char *dev, int *bus, int *id, int *lun)
 	devfile = g_strdup_printf ("/dev/%s", dev);
 	fd = open(devfile, O_RDONLY | O_NONBLOCK);
 	g_free (devfile);
-	
+
+	/* Avoid problems with Valgrind */
+	memset (&m_idlun, 1, sizeof (m_idlun));
+	*bus = *id = *lun = -1;
+
 	if (fd < 0) {
 		g_warning ("Failed to open cd device %s\n", dev);
 		return 0;
 	}
-    
-	if (ioctl (fd, SCSI_IOCTL_GET_BUS_NUMBER, bus) < 0) {
+
+	if (ioctl (fd, SCSI_IOCTL_GET_BUS_NUMBER, bus) < 0 || *bus < 0) {
 		g_warning ("Failed to get scsi bus nr\n");
 		close (fd);
 		return 0;
 	}
+
 	if (ioctl (fd, SCSI_IOCTL_GET_IDLUN, &m_idlun) < 0) {
 		g_warning ("Failed to get scsi id and lun\n");
 		close(fd);
@@ -324,7 +608,7 @@ get_cd_scsi_id (const char *dev, int *bus, int *id, int *lun)
 	}
 	*id = m_idlun.mux4 & 0xFF;
 	*lun = (m_idlun.mux4 >> 8)  & 0xFF;
-	
+
 	close(fd);
 	return 1;
 }
@@ -348,53 +632,39 @@ lookup_scsi_unit (int bus, int id, int lun,
 static int
 get_device_max_speed (char *id)
 {
-	int max_speed, i;
-	const char *argv[20]; /* Shouldn't need more than 20 arguments */
-	char *dev_str, *stdout_data, *speed;
+	int max_speed;
+	char *stdout_data, *speed;
 
 	max_speed = -1;
 
-	i = 0;
-	argv[i++] = "cdrecord";
-	argv[i++] = "-prcap";
-	dev_str = g_strdup_printf ("dev=%s", id);
-	argv[i++] = dev_str;
-	argv[i++] = NULL;
-
-	if (g_spawn_sync (NULL,
-			  (char **)argv,
-			  NULL,
-			  G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL,
-			  NULL, NULL,
-			  &stdout_data,
-			  NULL,
-			  NULL,
-			  NULL)) {
-		speed = strstr (stdout_data, "Maximum write speed in kB/s:");
-		if (speed != NULL) {
-			speed += strlen ("Maximum write speed in kB/s:");
-			max_speed = (int)floor (atol (speed) / 176.0 + 0.5);
-		} else {
-			speed = strstr (stdout_data, "Maximum write speed:");
-			if (speed != NULL) {
-				speed += strlen ("Maximum write speed:");
-				max_speed = (int)floor (atol (speed) / 176.0 + 0.5);
-			}
-		}
-		g_free (stdout_data);
+	stdout_data = cdrecord_get_stdout_for_id (id);
+	if (stdout_data == NULL) {
+		return max_speed;
 	}
 
-	g_free (dev_str);
+	speed = strstr (stdout_data, "Maximum write speed in kB/s:");
+	if (speed != NULL) {
+		speed += strlen ("Maximum write speed in kB/s:");
+		max_speed = (int)floor (atol (speed) / 176.0 + 0.5);
+	} else {
+		speed = strstr (stdout_data, "Maximum write speed:");
+		if (speed != NULL) {
+			speed += strlen ("Maximum write speed:");
+			max_speed = (int)floor (atol (speed) / 176.0 + 0.5);
+		}
+	}
+
+	g_free (stdout_data);
+
 	return max_speed;
 }
 
-
 static char *
 get_scsi_cd_name (int bus, int id, int lun, const char *dev,
 		  struct scsi_unit *scsi_units, int n_scsi_units)
 {
 	struct scsi_unit *scsi_unit;
-	
+
 	scsi_unit = lookup_scsi_unit (bus, id, lun, scsi_units, n_scsi_units);
 	if (scsi_unit == NULL) {
 		return g_strdup_printf (_("Unnamed SCSI CD-ROM (%s)"), dev);
@@ -575,7 +845,7 @@ linux_scan (gboolean recorder_only)
 	 * we need the oldstyle name to get device->display_name for ide.
 	 */
 	have_devfs = FALSE;
-	if (g_file_test ("/dev/cdroms", G_FILE_TEST_IS_DIR)) {
+	if (g_file_test ("/dev/.devfsd", G_FILE_TEST_EXISTS)) {
 		have_devfs = TRUE;
 	}
 	
@@ -622,7 +892,7 @@ linux_scan (gboolean recorder_only)
 	if (n_scsi_units > 0) {
 		/* open /dev/sg0 to force loading of the sg module if not loaded yet */
 		fd = open ("/dev/sg0", O_RDONLY);
-		if (fd != -1) {
+		if (fd >= 0) {
 			close (fd);
 		}
 		
@@ -633,7 +903,7 @@ linux_scan (gboolean recorder_only)
 			g_strfreev (devices);
 			return NULL;
 		}
-		
+
 		scsi_units = g_new0 (struct scsi_unit, n_scsi_units);
 		get_scsi_units (device_str, devices, scsi_units);
 
@@ -642,7 +912,7 @@ linux_scan (gboolean recorder_only)
 	} else {
 		scsi_units = NULL;
 	}
-	
+
 	for (i = 3; cdrom_info[i] != NULL; i++) {
 		if (g_str_has_prefix (cdrom_info[i], "Can write CD-R:")) {
 			p = cdrom_info[i] + strlen ("Can write CD-R:");
@@ -738,7 +1008,7 @@ linux_scan (gboolean recorder_only)
 		}
 		cdroms[i].display_name = cdrom_get_name (&cdroms[i],
 				scsi_units, n_scsi_units);
-		add_whitelist (&cdroms[i], scsi_units, n_scsi_units);
+		linux_add_whitelist (&cdroms[i], scsi_units, n_scsi_units);
 
 		if ((cdroms[i].can_write_cdr ||
 		    cdroms[i].can_write_cdrw ||
@@ -770,111 +1040,8 @@ linux_scan (gboolean recorder_only)
 
 	return cdroms_list;
 }
-#endif /* __linux__ */
 
-#ifdef __FreeBSD__
-static void
-get_cd_properties (char *id, int *max_rd_speed, int *max_wr_speed,
-	CDDriveType *type)
-{
-    	int i;
-	const char *argv[20];
-	char *dev_str, *stdout_data, *rd_speed, *wr_speed, *drive_cap;
-
-	*max_rd_speed = -1;
-	*max_wr_speed = -1;
-	*type = 0;
-
-	i = 0;
-	argv[i++] = "cdrecord";
-	argv[i++] = "-prcap";
-	dev_str = g_strdup_printf ("dev=%s", id);
-	argv[i++] = dev_str;
-	argv[i++] = NULL;
-
-	if (g_spawn_sync (NULL,
-			  (char **)argv,
-			  NULL,
-			  G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL,
-			  NULL, NULL,
-			  &stdout_data,
-			  NULL,
-			  NULL,
-			  NULL)) {
-		wr_speed = strstr (stdout_data, "Maximum write speed:");
-		if (wr_speed != NULL) {
-		    	char *tok;
-		    	wr_speed += strlen ("Maximum write speed:");
-			for (tok = strtok (wr_speed, " (),\t\n");
-				tok && strcmp (tok, "CD");
-				tok = strtok (NULL, " (),\t\n")) {}
-			tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
-			*max_wr_speed = atol (tok);
-		}
-		else {
-		    	wr_speed = strstr (stdout_data, "Maximum write speed in kB/s:");
-			if (wr_speed != NULL) {
-			    	char *tok;
-			    	wr_speed += strlen ("Maximum write speed in kB/s:");
-				for (tok = strtok (wr_speed, " (),\t\n");
-					tok && strcmp (tok, "CD");
-					tok = strtok (NULL, " (),\t\n")) {}
-				tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
-				*max_wr_speed = atol (tok);
-			}
-		}
-
-		rd_speed = strstr (stdout_data, "Maximum read  speed:");
-		if (rd_speed != NULL) {
-		    	char *tok;
-			rd_speed += strlen ("Maximum read  speed:");
-			for (tok = strtok (rd_speed, " (),\t\n"); 
-				tok && strcmp (tok, "CD"); 
-				tok = strtok (NULL, " (),\t\n")) {}
-			tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
-			*max_rd_speed = atol (tok);
-		}
-		else {
-		    	rd_speed = strstr (stdout_data, "Maximum read  speed in kB/s:");
-			if (rd_speed != NULL) {
-			    	char *tok;
-			    	rd_speed += strlen ("Maximum read  speed in kB/s:");
-				for (tok = strtok (rd_speed, " (),\t\n");
-					tok && strcmp (tok, "CD");
-					tok = strtok (NULL, " (),\t\n")) {}
-				tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
-				*max_rd_speed = atol (tok);
-			}
-		}
-		drive_cap = strstr (stdout_data, "Does write DVD-RAM media");
-		if (drive_cap != NULL) {
-		    	*type |= CDDRIVE_TYPE_DVD_RAM_RECORDER;
-		}
-		drive_cap = strstr (stdout_data, "Does read DVD-R media");
-		if (drive_cap != NULL) {
-		    	*type |= CDDRIVE_TYPE_DVD_RW_RECORDER;
-		}
-		drive_cap = strstr (stdout_data, "Does read DVD-ROM media");
-		if (drive_cap != NULL) {
-		    	*type |= CDDRIVE_TYPE_DVD_DRIVE;
-		}
-		drive_cap = strstr (stdout_data, "Does write CD-RW media");
-		if (drive_cap != NULL) {
-		    	*type |= CDDRIVE_TYPE_CDRW_RECORDER;
-		}
-		drive_cap = strstr (stdout_data, "Does write CD-R media");
-		if (drive_cap != NULL) {
-		    	*type |= CDDRIVE_TYPE_CD_RECORDER;
-		}
-		drive_cap = strstr (stdout_data, "Does read CD-R media");
-		if (drive_cap != NULL) {
-		    	*type |= CDDRIVE_TYPE_CD_DRIVE;
-		}
-		g_free (stdout_data);
-	}
-
-	g_free (dev_str);
-}
+#elif defined (__FreeBSD__)
 
 static GList *
 freebsd_scan (gboolean recorder_only)
@@ -937,8 +1104,7 @@ freebsd_scan (gboolean recorder_only)
 			}
 
 			cdroms_list = g_list_append (cdroms_list, cdrom);
-		}
-		else {
+		} else {
 		    	cd_drive_free (cdrom);
 		}
 
@@ -950,30 +1116,147 @@ freebsd_scan (gboolean recorder_only)
 
 	return cdroms_list;
 }
-#endif /* __FreeBSD__ */
+
+#else
+
+
+
+static char *
+cdrecord_scan_get_stdout (void)
+{
+	int max_speed, i;
+	const char *argv[20]; /* Shouldn't need more than 20 arguments */
+	char *stdout_data;
+
+	max_speed = -1;
+
+	i = 0;
+	argv[i++] = "cdrecord";
+	argv[i++] = "-scanbus";
+	argv[i++] = NULL;
+
+	if (g_spawn_sync (NULL,
+				(char **)argv,
+				NULL,
+				G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL,
+				NULL, NULL,
+				&stdout_data,
+				NULL,
+				NULL,
+				NULL)) {
+		return stdout_data;
+	}
+
+	return NULL;
+}
+
+#define DEFAULT_SPEED 2
+
+static GList *
+cdrecord_scan (gboolean recorder_only)
+{
+	GList *cdroms_list;
+	CDDrive *cdrom;
+	char *stdout_data, **lines, vendor[9], model[17];
+	int i, bus, id, lun, index;
+
+	cdroms_list = NULL;
+
+	stdout_data = cdrecord_scan_get_stdout ();
+	if (stdout_data == NULL) {
+		return cdroms_list;
+	}
+
+	lines = g_strsplit (stdout_data, "\n", 0);
+	g_free (stdout_data);
+
+	for (i = 0; lines[i] != NULL; i++) {
+		if (sscanf (lines[i], "\t%d,%d,%d\t  %d) '%8c' '%16c'",
+					&bus, &id, &lun, &index,
+					vendor, model) != 6) {
+			continue;
+		}
+
+		vendor[8] = '\0'; model[16] = '\0';
+
+		cdrom = g_new0 (CDDrive, 1);
+		cdrom->display_name = g_strdup_printf ("%s - %s",
+				g_strstrip (vendor), g_strstrip (model));
+		cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d", bus, id, lun);
+		/* FIXME we don't have any way to guess the real device
+		 * from the info we get from CDRecord */
+		cdrom->device = g_strdup_printf ("/dev/%d", index);
+		get_cd_properties (cdrom->cdrecord_id,
+			&(cdrom->max_speed_read),
+			&(cdrom->max_speed_write),
+			&(cdrom->type));
+		add_whitelist (cdrom);
+		if (cdrom->type & CDDRIVE_TYPE_CD_RECORDER
+				|| cdrom->type & CDDRIVE_TYPE_CDRW_RECORDER
+				|| cdrom->type & CDDRIVE_TYPE_DVD_RAM_RECORDER
+				|| cdrom->type & CDDRIVE_TYPE_DVD_RW_RECORDER
+				|| !recorder_only) {
+
+			if (cdrom->max_speed_read == -1) {
+		    		cdrom->max_speed_read = DEFAULT_SPEED;
+			}
+			if (cdrom->max_speed_write == -1) {
+			    	cdrom->max_speed_write = DEFAULT_SPEED;
+			}
+
+			if (cdrom->type & CDDRIVE_TYPE_DVD_DRIVE) {
+				add_dvd_plus (cdrom);
+			}
+
+			cdroms_list = g_list_append (cdroms_list, cdrom);
+		} else {
+		    	cd_drive_free (cdrom);
+		}
+	}
+
+	g_strfreev (lines);
+
+	return cdroms_list;
+}
+
+#endif
+
+CDDrive *
+cd_drive_get_file_image (void)
+{
+	CDDrive *cdrom;
+
+	cdrom = g_new0 (CDDrive, 1);
+	cdrom->display_name = g_strdup (_("File image"));
+	cdrom->max_speed_read = 0;
+	cdrom->max_speed_write = 0;
+	cdrom->type = CDDRIVE_TYPE_FILE;
+
+	return cdrom;
+}
 
 GList *
 scan_for_cdroms (gboolean recorder_only, gboolean add_image)
 {
-	CDDrive *cdrom;
 	GList *cdroms = NULL;
 
-#ifdef __linux__
-	cdroms = linux_scan (recorder_only);
+#ifdef USE_HAL
+	cdroms = hal_scan (recorder_only);
 #endif
 
-#ifdef __FreeBSD__
-	cdroms = freebsd_scan (recorder_only);
+	if (cdroms == NULL) {
+#if defined (__linux__)
+		cdroms = linux_scan (recorder_only);
+#elif defined (__FreeBSD__)
+		cdroms = freebsd_scan (recorder_only);
+#else
+		cdroms = cdrecord_scan (recorder_only);
 #endif
+	}
 
 	if (add_image) {
-		/* File */
-		cdrom = g_new0 (CDDrive, 1);
-		cdrom->display_name = g_strdup (_("File image"));
-		cdrom->max_speed_read = 0;
-		cdrom->max_speed_write = 0;
-		cdrom->type = CDDRIVE_TYPE_FILE;
-
+		CDDrive *cdrom;
+		cdrom = cd_drive_get_file_image ();
 		cdroms = g_list_append (cdroms, cdrom);
 	}
 
@@ -991,13 +1274,3 @@ cd_drive_free (CDDrive *drive)
 	g_free (drive);
 }
 
-CDMediaType
-guess_media_type (const char *device_path)
-{
-#if defined(__linux__) || defined(__FreeBSD__)
-	return linux_bsd_media_type (device_path);
-#else
-	return CD_MEDIA_TYPE_UNKNOWN;
-#endif
-}
-
